import pyspark.sql.functions as F
from pyspark.sql.window import Window

def filter_pivot_bbga(df,variabelen,meest_recente_jaar = True):
    '''
    Filtert het Basisbestand Gebieden Amsterdam (BBGA) op variabelen en het meeste recente jaar of niet.
    args: 
    df: de BBGA in de vorm van een PySpark DataFrame
    variabelen: een lijst van variabelen of een dictionary met variabelen als keys en aliassen als values
    meest_recente_jaar: boolean, standaard True
    return: een PySpark DataFrame met de gewenste variabelen en het meest recente jaar tenzij anders aangegeve
    '''
    # Filter op variabelen en alias indien dictionary
    if isinstance(variabelen, dict):
        # Maak een mapping expression om kolomwaarden te hernoemen (row-based equivalent van kolommen aliasen)
        mapping_expr = F.create_map(*[F.lit(x) for x in sum(variabelen.items(), ())])

        # Filter het df met een where clause om alleen records met de gewenste indicator over te houden
        df_gefilterd = df.filter(F.col("Indicatordefinitieid").isin(list(variabelen.keys())))

        # Pas de mapping toe op de bestaande kolom
        df_gefilterd = df_gefilterd.withColumn("Indicatordefinitieid", mapping_expr[F.col("Indicatordefinitieid")])

    elif isinstance(variabelen, list):
        df_gefilterd = df.filter(F.col("Indicatordefinitieid").isin(variabelen))

    else:
        raise TypeError(
            "variabelen moet een list of dict zijn"
        )

    # Filter op het meest recente jaar 
    df_gefilterd_jaar = df_gefilterd.withColumn("Jaar", F.col("Jaar").cast("int"))
    df_gefilterd_jaar = df_gefilterd_jaar.withColumn("Waarde", F.col("Waarde").cast("double"))

    # Zet binnen window van gebied en indicator een row number op descending Jaar en neem rij 1 voor het meest recente jaar, groupby alleen op gebied  
    if meest_recente_jaar:
        window_spec = (
            Window
            .partitionBy("Gebiedcode15", "Indicatordefinitieid")
            .orderBy(F.col("Jaar").desc())
        )

        df_gefilterd_jaar = (
            df_gefilterd_jaar
            .withColumn("row_number", F.row_number().over(window_spec))
            .filter(F.col("row_number") == 1)
            .drop("row_number")
        )

        # Pivot het gefilterde df
        df_pivot = (
        df_gefilterd_jaar
        .groupBy("Gebiedcode15")
        .pivot("Indicatordefinitieid")
        .agg(F.first("Waarde"))
        )

    # Indien meerdere jaren gewenst zijn, pivot en groupby op zowel gebied als jaar (anders wordt jaar gedropt en een random eerste waarde gepakt door First)
    else:
        # Pivot het gefilterde df
        df_pivot = (
        df_gefilterd_jaar
        .groupBy("Gebiedcode15","Jaar")
        .pivot("Indicatordefinitieid")
        .agg(F.first("Waarde"))
        )    
      
    return df_pivot



# Databricks notebook source

# Import necessary Databricks runtime objects
from databricks.sdk.runtime import *
from pyspark.sql import SparkSession


def bronanalyse(spark: SparkSession, catalog: str, schemas: list, trefwoorden_kolomnamen: list, counts = False, waarden = False):

    """
    Labelt de rijen van een DataFrame op basis van categorieën en matchende strings.

    Args:
        catalog (str): De naam van de catalog waar je de analyse op wil doen
        schemas (list): De namen van een of meerdere schema's waar je de analyse op wil doen
        trefwoorden_kolomnamen (list): Een of meerdere trefwoorden om kolomnamen op te selecteren
        counts (bool): De optie om een count en count distinct per kolom op te halen.
        waarden (bool): De optie om de eerste 5 unieke waarden per kolom op te halen.

    Returns:
        pd.Dataframe: een dataframe met tabel- en kolomnamen en bijbehorende counts en waarden.
    """
    
    #selectie van kolommen uit information_schema die we gebruiken voor bronanalyse
    kolommen_df = ['table_catalog','table_schema','table_name','column_name','data_type']

    #haal information schema op 
    df = spark.table(f'{catalog}.information_schema.columns').select(*kolommen_df)

    #filter information_schema tot relevante scheman
    schemas = '|'.join(schemas)
    df = df.where(df['table_schema'].contains(schemas))
                     
    #filter information_schema tot relevante kolomnamen
    kolomnamen = '|'.join(trefwoorden_kolomnamen)
    df = df.where(df['column_name'].contains(kolomnamen))

    #naar pandas
    df = df.toPandas()

    # bereken counts indien true
    if counts == True:
        #maak queries
        #count
        df['query_count'] = 'SELECT COUNT('+df['column_name']+') FROM '+df['table_catalog']+'.'+df['table_schema']+'.'+df['table_name']
        #count distinct
        df['query_count_distinct'] = 'SELECT COUNT(DISTINCT('+df['column_name']+')) FROM '+df['table_catalog']+'.'+df['table_schema']+'.'+df['table_name']
        #query data
        df['count'] = df['query_count'].apply(lambda x: spark.sql(x).first()[0])
        df['count_distinct'] = df['query_count_distinct'].apply(lambda x: spark.sql(x).first()[0])
    
    if waarden == True:
        #maak query
        df['query_waarden'] = 'SELECT DISTINCT('+df['column_name']+') FROM '+df['table_catalog']+'.'+df['table_schema']+'.'+df['table_name']+' LIMIT 5'
        #query data
        df['5_eerste_unieke_waarden'] = df['query_waarden'].apply(lambda x: ';'.join ([str(row[0]) for row in spark.sql(x).collect()]))
    
    #schoon dataframe op door queries te verwijderen
    if counts == True or waarden == True:
        for column in df.columns:
            if 'query' in column:
                df = df.drop(column,axis=1)
            else:
                pass
    return df



from pyspark.sql.functions import col, lit
from pyspark.sql.utils import AnalysisException # Importeer de specifieke uitzondering voor SQL-fouten
import sys

# 
# Hier staan generieke functies die gebruikt worden om aanpasingen te doen aan tabellen en kollommen via eenmalige
# scripts. Deze functies zijn ontworpen zodat dit door de hele OTAP straat op uitgeteste en gecontrolleerde wijze kan plaatsvinden
#
#>> FUNCTIES IN DEZE FILE ZIJN:
#
# 1 hernoem_kolommen_met_delta_mapping
# 2 voeg_kolommen_toe
# 3 _get_default_literal
# 4 vervang_tabel_met_tijdelijke_tabel
# 5 voeg_kolommen_toe_op_positie
# 6 pas_datatypes_aan
# 7 verwijder_kolommen_uit_delta_tabel
# 8 pas_kolomvolgorde_aan
# 9 update_table_from_other_table
#

#
#> 1 HERNOEM KOLLOMMEN MET DELTA MAPPING
#

def hernoem_kolommen_met_delta_mapping(spark, catalog, target_schema, tabel_kolom_rename_mapping):
    """
    Hernoemt kolommen in meerdere tabellen en schakelt Delta column mapping in.
    Geeft een foutmelding als een kolom niet bestaat, maar blijft de rest van de kolommen verwerken.

    Parameters:
    - spark: De SparkSession (vereist in Databricks).
    - catalog: De naam van de catalogus.
    - target_schema: De naam van het doel-schema.
    - tabel_kolom_rename_mapping: Een dictionary waarin de sleutels de tabelnamen zijn
      en de waarden dictionaries zijn met oude kolomnamen als sleutels en nieuwe kolomnamen als waarden.
    """
    for originele_tabel, kolom_mapping in tabel_kolom_rename_mapping.items():
        # Definieer de volledige naam van de tabel
        volledige_tabel_naam = f"{catalog}.{target_schema}.{originele_tabel}"
        
        print(f"\nBezig met tabel: {volledige_tabel_naam}")

        # 1. Schakel Delta column mapping in voor de tabel
        try:
            spark.sql(f"""
                ALTER TABLE {volledige_tabel_naam} SET TBLPROPERTIES (
                    'delta.columnMapping.mode' = 'name',
                    'delta.minReaderVersion' = '2',
                    'delta.minWriterVersion' = '5'
                )
            """)
            print(f"✅ Delta column mapping ingeschakeld voor: {volledige_tabel_naam}")
        except AnalysisException as e:
            # Vang de AnalysisException op als de tabel niet bestaat of een andere fout optreedt bij TBLPROPERTIES
            print(f"❌ Fout bij het inschakelen van Delta column mapping voor {volledige_tabel_naam}: {e}")
            continue # Ga naar de volgende tabel als er een fout is bij het inschakelen van de mapping.

        # 2. Hernoem de kolommen in de tabel
        for oude_kolom, nieuwe_kolom in kolom_mapping.items():
            try:
                spark.sql(f"""
                    ALTER TABLE {volledige_tabel_naam} RENAME COLUMN `{oude_kolom}` TO `{nieuwe_kolom}`
                """)
                print(f"   ✅ Kolom hernoemd: **`{oude_kolom}`** -> **`{nieuwe_kolom}`**")
            except AnalysisException as e:
                # Vang de AnalysisException op die optreedt als de kolom niet bestaat
                # of een andere SQL-gerelateerde fout
                foutboodschap = getattr(e, 'errorMessage', str(e)) 
            
                print(f"   ⚠️ FOUT: Kolom **`{oude_kolom}`** kon NIET worden hernoemd in tabel **`{volledige_tabel_naam}`** naar **`{nieuwe_kolom}`**.")
                print(f"   Reden: {foutboodschap.split(':')[0]}") # Print een deel van de foutboodschap voor beknoptheid
                # De lus gaat hier verder naar de volgende kolom in de tabel
    
    print("\nAlle hernoemingen zijn verwerkt. Controleer de ⚠️ FOUT berichten voor mislukte renamen.")
    

#
#> 2 VOEG KOLLOMMEN TOE AAN DELTA TABEL
#

def voeg_kolommen_toe(spark, catalog, target_schema, kolom_toevoeg_mapping):
    """
    Voeg nieuwe kolommen toe aan Delta-tabellen.

    Parameters:
    - spark: De SparkSession (vereist in Databricks).
    - catalog: De naam van de catalogus.
    - target_schema: De naam van het doel-schema.
    - kolom_toevoeg_mapping: Een dictionary waarin de sleutels de tabelnamen zijn
      en de waarden dictionaries zijn met kolomnamen als sleutels en een tuple (datatype, nullable) als waarden.
    """
    for tabel, kolom_mapping in kolom_toevoeg_mapping.items():
        volledige_tabel_naam = f"{catalog}.{target_schema}.{tabel}"
        
        for kolom, (datatype, nullable) in kolom_mapping.items():
            # Bepaal of de kolom nullable is
            nullable_str = "NULL" if nullable else "NOT NULL"
            
            # SQL-query om de kolom toe te voegen
            spark.sql(f"""
                ALTER TABLE {volledige_tabel_naam}
                ADD COLUMNS ({kolom} {datatype} {nullable_str})
            """)
            print(f"Kolom toegevoegd: {kolom} ({datatype}, {nullable_str}) in tabel {volledige_tabel_naam}")

    print("Alle nieuwe kolommen zijn succesvol toegevoegd.")

#
#> 3 GET DEFAULT LITERAL
#

def _get_default_literal(datatype: str):
    """
    Bepaalt de standaardwaarde (literal) op basis van het datatype voor 
    nieuwe NOT NULL kolommen. Dit is cruciaal om te voorkomen dat 
    ALTER TABLE ... SET NOT NULL faalt.
    """
    datatype = datatype.upper()
    if datatype in ("STRING", "VARCHAR"):
        return lit("")
    elif datatype in ("INT", "INTEGER", "BIGINT", "SMALLINT"):
        return lit(0)
    elif datatype in ("DOUBLE", "FLOAT", "DECIMAL"):
        return lit(0.0)
    elif datatype == "BOOLEAN":
        return lit(False)
    elif datatype == "DATE":
        return lit("1900-01-01").cast("date") 
    elif datatype == "TIMESTAMP":
        # Gebruik de epoch datum/tijd als veilige standaard
        return lit("1970-01-01 00:00:00").cast("timestamp")
    else:
        # Voor onbekende of complexe types, val terug op NULL (waarschuwing: DDL zal falen!)
        print(f"WAARSCHUWING: Onbekend datatype '{datatype}' voor NOT NULL kolom. Gebruikt lit(None).")
        return lit(None)

#
#> 4 VERVANG TABEL MET TIJDELIJKE TABEL
#

def vervang_tabel_met_tijdelijke_tabel(
    spark, 
    oude_tabel_pad: str, 
    nieuwe_tabel_pad: str, 
    nieuwe_df, 
    nieuwe_kolom_definitie: dict
):
    """
    Vervangt een bestaande tabel door een nieuwe tijdelijke tabel en past NOT NULL constraints toe.
    Deze functie is aangepast om de DDL (ALTER TABLE) uit te voeren.
    """
    # Stap 1: Schrijf de nieuwe data naar de tijdelijke tabel
    print(f"Schrijft data naar tijdelijke tabel: {nieuwe_tabel_pad}")
    nieuwe_df.write.format("delta").mode("overwrite").saveAsTable(nieuwe_tabel_pad)
    
    # Optionele Stap: Controleer het aantal rijen
    try:
        # Probeer het aantal rijen te vergelijken (kan falen als oude_tabel_pad niet bestaat)
        originele_rijen_aantal = spark.table(oude_tabel_pad).count()
        nieuwe_rijen_aantal = spark.table(nieuwe_tabel_pad).count()
        if nieuwe_rijen_aantal != originele_rijen_aantal:
             raise ValueError(f"Aantal rijen in de nieuwe tabel ({nieuwe_rijen_aantal}) komt niet overeen met het aantal rijen in de originele tabel ({originele_rijen_aantal}).")
        print(f"Data succesvol overgezet naar tijdelijke tabel: {nieuwe_tabel_pad}. Rijentelling is correct ({nieuwe_rijen_aantal}).")
    except Exception as e:
        print(f"INFO: Kon rijen niet controleren voor {oude_tabel_pad} (fout: {e}). Gaat door.")

    # Stap 2: Verwijder de originele tabel
    print(f"Verwijdert originele tabel {oude_tabel_pad}...")
    spark.sql(f"DROP TABLE IF EXISTS {oude_tabel_pad}")

    # Stap 3: Hernoem de tijdelijke tabel naar de originele tabel
    print(f"Hernoemt tijdelijke tabel {nieuwe_tabel_pad} naar {oude_tabel_pad}")
    spark.sql(f"ALTER TABLE {nieuwe_tabel_pad} RENAME TO {oude_tabel_pad}")
    
    # Stap 4: Pas de NOT NULL constraints toe op de nieuwe kolommen
    print("Controleert op nieuwe kolommen met NOT NULL constraint...")
    for naam, (datatype, positie, nullable) in nieuwe_kolom_definitie.items():
        if not nullable:
            # Dit DDL-commando werkt op Delta Lake en de meeste Spark SQL-implementaties.
            alter_sql = f"ALTER TABLE {oude_tabel_pad} ALTER COLUMN {naam} SET NOT NULL"
            
            print(f"-> Voert DDL uit voor NOT NULL op kolom {naam}: {alter_sql}")
            try:
                spark.sql(alter_sql)
            except Exception as e:
                # Als de DDL faalt (bijv. omdat de standaardwaarde niet correct was, of de syntax niet wordt ondersteund)
                print(f"FOUT: Kon NOT NULL niet toepassen op kolom {naam} in {oude_tabel_pad}. Fout: {e}")

    # Stap 5: Vernieuw de metadata
    print(f"Vernieuwt de catalogus metadata voor tabel {oude_tabel_pad}...")
    spark.catalog.refreshTable(oude_tabel_pad)

    print(f"Tabel {oude_tabel_pad} succesvol bijgewerkt.")

#
#> 5 VOEG KOLLOMMEN TOE AAN DELTA TABEL OP POSITIE
#

def voeg_kolommen_toe_op_positie(spark, catalog: str, target_schema: str, kolom_toevoeg_mapping: dict):
    """
    Voegt nieuwe kolommen toe aan Delta-tabellen op een specifieke positie en dwingt 
    NOT NULL constraints af via DDL.
    """
    for tabel, kolom_mapping in kolom_toevoeg_mapping.items():
        # Definieer de volledige naam van de tabellen
        oude_tabel_pad = f"{catalog}.{target_schema}.{tabel}"
        nieuwe_tabel_pad = f"{oude_tabel_pad}_temp_met_nieuwe_kolommen"

        print(f"\n--- Start verwerking voor tabel: {oude_tabel_pad} ---")

        try:
            # Lees de data van de originele tabel in een DataFrame
            df = spark.table(oude_tabel_pad)
            originele_kolommen = [c.name for c in df.schema]
            
            # Bepaal de definitieve kolomvolgorde en de nieuwe kolommen
            nieuwe_kolom_definitie = {}
            for naam, datatype, positie, nullable in kolom_mapping:
                # Sla de definitie op voor later gebruik in de DDL stap
                nieuwe_kolom_definitie[naam] = (datatype, positie, nullable)

            definitieve_kolom_volgorde = []
            
            # Voeg de nieuwe kolommen in op de juiste positie
            for kolom_naam in originele_kolommen:
                # Kolommen VÓÓR de huidige kolom
                for nieuwe_kolom, (datatype, positie, nullable) in nieuwe_kolom_definitie.items():
                    if positie == kolom_naam:
                        definitieve_kolom_volgorde.append((nieuwe_kolom, datatype, nullable))
                
                # De Oude kolom zelf
                definitieve_kolom_volgorde.append((kolom_naam, None, None))

            # Voeg kolommen toe die aan het einde moeten komen (positie is None)
            for nieuwe_kolom, (datatype, positie, nullable) in nieuwe_kolom_definitie.items():
                if positie is None:
                    definitieve_kolom_volgorde.append((nieuwe_kolom, datatype, nullable))

            # Bouw de selectiekolommen met de juiste volgorde, types en (standaard)waarden
            kolommen_om_te_selecteren = []
            for naam, datatype, nullable in definitieve_kolom_volgorde:
                if datatype is None:  # Oude kolom
                    kolommen_om_te_selecteren.append(col(naam))
                else:  # Nieuwe kolom
                    if nullable:
                        # Nieuwe kolom is NULLABLE: voeg een NULL toe
                        kolommen_om_te_selecteren.append(lit(None).cast(datatype).alias(naam))
                    else:
                        # Nieuwe kolom is NOT NULL: voeg een STANDAARDWAARDE toe
                        standaard_waarde_lit = _get_default_literal(datatype)
                        kolommen_om_te_selecteren.append(standaard_waarde_lit.cast(datatype).alias(naam))

            # Selecteer de kolommen met de bijgewerkte datatypes in een nieuwe DataFrame
            nieuwe_df = df.select(kolommen_om_te_selecteren)
            
            print(f"Nieuw schema voor {tabel}:")
            nieuwe_df.printSchema()
            
            # Gebruik de helperfunctie om de tabel te vervangen en DDL toe te passen
            vervang_tabel_met_tijdelijke_tabel(
                spark, 
                oude_tabel_pad, 
                nieuwe_tabel_pad, 
                nieuwe_df, # Geef de nieuwe DF mee
                nieuwe_kolom_definitie # Geef de mapping mee voor DDL
            )

        except Exception as e:
            print(f"Fout opgetreden bij verwerking van tabel {oude_tabel_pad}: {e}", file=sys.stderr)
            raise

    print("\nAlle nieuwe kolommen zijn succesvol toegevoegd op de gewenste posities.")

#
# 6 PAS DATATYPES AAN VAN DELTA TABEL
#

def pas_datatypes_aan(spark, catalog, target_schema, kolom_data_type_mapping):
    """
    Wijzigt alleen de datatypes van opgegeven kolommen in Delta-tabellen.
    Kolommen die beginnen met 'sid_' worden uitgesloten van de gegevensoverdracht.
    Andere kolomeigenschappen, zoals AUTO INCREMENT, blijven behouden.

    Parameters:
    - spark: De SparkSession (vereist in Databricks).
    - catalog: De naam van de catalogus.
    - target_schema: De naam van het doel-schema.
    - kolom_data_type_mapping: Een dictionary waarin de sleutels de tabelnamen zijn
      en de waarden dictionaries zijn met kolomnamen als sleutels en hun nieuwe datatypes als waarden.
    """
    for originele_tabel, kolom_mapping in kolom_data_type_mapping.items():
        # Definieer de volledige namen van de tabellen
        oude_tabel_pad = f"{catalog}.{target_schema}.{originele_tabel}"
        nieuwe_tabel_pad = f"{oude_tabel_pad}_temp_nieuwe_datatypes"

        print(f"Start verwerking voor tabel: {oude_tabel_pad}")

        try:
            # Lees de data van de originele tabel in een DataFrame
            df = spark.table(oude_tabel_pad)

            # Bepaal welke kolommen behouden moeten worden en welke van datatype moeten veranderen
            kolommen_om_te_selecteren = []
            for kolom_naam in df.columns:
                # Kolommen die beginnen met 'sid_' worden overgeslagen
                if kolom_naam.lower().startswith("sid_"):
                    print(f"Kolom '{kolom_naam}' wordt overgeslagen (SID-kolom).")
                    kolommen_om_te_selecteren.append(col(kolom_naam))
                elif kolom_naam in kolom_mapping:
                    # Cast de kolom naar het nieuwe datatype
                    nieuw_type = kolom_mapping[kolom_naam]
                    print(f"Kolom '{kolom_naam}' wordt gecast naar '{nieuw_type}'.")
                    kolommen_om_te_selecteren.append(col(kolom_naam).cast(nieuw_type))
                else:
                    # Behoud de kolom zoals deze is
                    kolommen_om_te_selecteren.append(col(kolom_naam))

            # Selecteer de kolommen met de bijgewerkte datatypes in een nieuwe DataFrame
            nieuwe_df = df.select(*kolommen_om_te_selecteren)
            
           # Gebruik de nieuwe functie om de tabel te vervangen
            vervang_tabel_met_tijdelijke_tabel(spark, oude_tabel_pad, nieuwe_tabel_pad)

        except Exception as e:
            print(f"Fout opgetreden bij verwerking van tabel {oude_tabel_pad}: {e}")
            raise

    print("Alle opgegeven datatypes zijn succesvol aangepast.")

#
# 7 VERWIJDER KOLLOMMEN UIT DELTA TABEL
#

def verwijder_kolommen_uit_delta_tabel(spark, catalog: str, target_schema: str, kolom_verwijder_mapping: dict):
    """
    Verwijdert kolommen uit Delta-tabellen door een back-up te maken en de tabel opnieuw te creëren,
    waarna de nieuwe opgeschoonde tabel de oorspronkelijke tabel vervangt via hernoeming.
    
    Belangrijkste wijziging: Als de originele tabel een kolom bevatte die begint met 'sid_',
    dan wordt deze in de nieuwe tabel opnieuw aangemaakt als een automatisch gegenereerde 
    BIGINT IDENTITY kolom, via de correcte DDL-methode (CREATE TABLE + INSERT INTO).
    
    Belangrijk: Alle kolomnamen die speciale tekens bevatten (zoals '@') worden nu 
    automatisch geciteerd met backticks (`) in de gegenereerde SQL-statements om 
    Syntax Errors te voorkomen.

    Parameters:
    - spark: De SparkSession.
    - catalog: De naam van de catalogus.
    - target_schema: De naam van het doel-schema.
    - kolom_verwijder_mapping: Een dictionary waarin de sleutels de tabelnamen zijn
      en de waarden lijsten zijn van kolommen die moeten worden verwijderd.
    """
    
    # Gebruik een set voor snellere lookup van te verwijderen kolommen (niet direct gebruikt in de loop, maar goed voor opschoning)
    # te_verwijderen_kolommen_totaal = set()
    # for kolommen in kolom_verwijder_mapping.values():
    #     te_verwijderen_kolommen_totaal.update([k.lower() for k in kolommen]) # De totale set is niet langer nodig, we verwerken per tabel

    for originele_tabel_naam, kolommen_verwijderen in kolom_verwijder_mapping.items():
        # Definieer de volledige namen van de tabellen
        originele_tabel = f"{catalog}.{target_schema}.{originele_tabel_naam}"
        backup_tabel = f"{originele_tabel}_backup"
        nieuwe_tabel = f"{originele_tabel}_nw" 

        try:
            print(f"--- Start verwerking tabel: {originele_tabel} ---")

            # 1. Bepaal de kolommen van de originele tabel en check op sid_
            try:
                # Laad de tabel met een beperking (limit 0) om het schema snel te verkrijgen
                df_origineel = spark.table(originele_tabel) 
            except AnalysisException:
                print(f"Tabel {originele_tabel} bestaat niet. Overslaan.")
                continue

            originele_kolommen = df_origineel.columns
            # Maak een set van te verwijderen kolommen (in lowercase) voor snelle lookup
            kolommen_verwijderen_set = {k.lower() for k in kolommen_verwijderen}
            
            # Zoek de eerste 'sid_' kolom (er mag slechts één identity kolom zijn)
            sid_kolom_naam = next((
                kolom for kolom in originele_kolommen 
                if kolom.lower().startswith("sid_")
            ), None)

            if sid_kolom_naam:
                print(f"Opmerking: Identity kolom '{sid_kolom_naam}' gevonden. Deze wordt opnieuw aangemaakt in de nieuwe tabel met IDENTITY eigenschappen.")
            
            # 2. Bepaal de kolommen die BEHOUDEN moeten blijven in de data-overdracht
            # We sluiten alle 'sid_' kolommen en de kolommen uit de lijst uit.
            kolommen_behouden = [
                kolom for kolom in originele_kolommen
                if not kolom.lower().startswith("sid_") and kolom.lower() not in kolommen_verwijderen_set
            ]

            # 3. Selecteer de data (zonder de verwijderde kolommen en zonder sid_)
            df_gefilterd = df_origineel.select(*kolommen_behouden)
            print(f"Kolommen behouden voor data-overdracht: {kolommen_behouden}")

            # 4. Schrijf de gefilterde data naar de back-up tabel (VEILIGHEIDSSTAP)
            spark.sql(f"DROP TABLE IF EXISTS {backup_tabel}") # Zorg dat back-up schoon is
            df_gefilterd.write.format("delta").mode("overwrite").saveAsTable(backup_tabel)
            
            # Count checks
            # Gebruik de originele DataFrame count voor vergelijking (hoewel een table count ook werkt)
            # Voor een grotere robuustheid, tellen we de rijen na het schrijven:
            backup_rijen_aantal = spark.table(backup_tabel).count()
            originele_rijen_aantal = df_origineel.count() 

            if backup_rijen_aantal != originele_rijen_aantal:
                raise ValueError(f"Back-up tabel {backup_tabel} heeft niet hetzelfde aantal rijen ({backup_rijen_aantal}) als de originele tabel ({originele_rijen_aantal}).")

            print(f"Back-up tabel succesvol aangemaakt en gevuld met {backup_rijen_aantal} rijen.")

            # 5. Verwijder de oude NIEUWE tabelstructuur
            spark.sql(f"DROP TABLE IF EXISTS {nieuwe_tabel}")

            if sid_kolom_naam:
                # --- IDENTITY KOLOM PAD: DDL + INSERT ---
                
                # 5a. Bepaal het schema van de data kolommen
                # CRUCIALE WIJZIGING 1: Voeg backticks toe rond de veldnamen voor DDL.
                data_kolom_definities = ", ".join(
                    f"`{field.name}` {field.dataType.simpleString()}" 
                    for field in df_gefilterd.schema.fields
                )
                
                # 5b. Definieer de Identity kolom en het volledige DDL
                identity_def = f"`{sid_kolom_naam}` BIGINT GENERATED ALWAYS AS IDENTITY (START WITH 1 INCREMENT BY 1)"
                
                full_ddl = f"""
                    CREATE TABLE {nieuwe_tabel} (
                        {identity_def}, {data_kolom_definities}
                    )
                    USING DELTA
                """
                # De fout treedt hier op als de kolomnamen in data_kolom_definities niet gequote zijn.
                spark.sql(full_ddl)
                print(f"Nieuwe tabelstructuur met IDENTITY-kolom aangemaakt: {nieuwe_tabel}")

                # 5c. Voer de data-insert uit vanuit de back-up
                # CRUCIALE WIJZIGING 2: Voeg backticks toe rond de kolomnamen voor DML.
                kolommen_invoegen_gequote = ", ".join(f"`{k}`" for k in kolommen_behouden)
                
                # De insert vult alleen de datavelden, de sid_kolom wordt automatisch gegenereerd
                insert_sql = f"INSERT INTO {nieuwe_tabel} ({kolommen_invoegen_gequote}) SELECT {kolommen_invoegen_gequote} FROM {backup_tabel}"
                # De fout treedt ook hier op als de kolomnamen in kolommen_invoegen_gequote niet gequote zijn.
                spark.sql(insert_sql)
                
                print(f"Data succesvol ingevoegd in {nieuwe_tabel}, IDENTITY-kolom automatisch gevuld.")

            else:
                # --- GEEN IDENTITY KOLOM PAD: saveAsTable (Oorspronkelijke methode) ---
                
                # saveAsTable zal de nieuwe tabel AANMAKEN met het correcte, gefilterde schema.
                df_gefilterd.write.format("delta").mode("overwrite").saveAsTable(nieuwe_tabel)
                print(f"Nieuwe tabel succesvol aangemaakt en gevuld met saveAsTable: {nieuwe_tabel}")
            
            # Count checks (nu voor beide paden)
            nieuwe_rijen_aantal = spark.table(nieuwe_tabel).count()

            if nieuwe_rijen_aantal != backup_rijen_aantal:
                raise ValueError(f"Nieuwe tabel {nieuwe_tabel} heeft na de schrijfactie een onjuist aantal rijen ({nieuwe_rijen_aantal}). Back-up is beschikbaar via {backup_tabel}.")

            # --- STAPPEN 6, 7, 8: VERVANGING EN OPSCHONING ---

            # 6. Verwijder de ORIGINELE tabel
            spark.sql(f"DROP TABLE IF EXISTS {originele_tabel}")
            print(f"Originele tabel definitief verwijderd: {originele_tabel}")
            
            # 7. Hernoem de NIEUWE tabel naar de ORIGINELE naam
            spark.sql(f"ALTER TABLE {nieuwe_tabel} RENAME TO {originele_tabel}")
            print(f"Tabel {nieuwe_tabel} hernoemd naar {originele_tabel}. Migratie voltooid.")

            # 8. Verwijder de back-up tabel
            spark.sql(f"DROP TABLE IF EXISTS {backup_tabel}")
            print(f"Back-up tabel verwijderd: {backup_tabel}")
            
        except Exception as e:
            # Foutafhandeling
            print(f"!!! FOUT opgetreden bij verwerking van tabel {originele_tabel}: {e}")
            print(f"Belangrijk: De back-up tabel {backup_tabel} is BEWAARD en bevat de data ZONDER de te verwijderen kolommen. Handmatig herstel is mogelijk.")
            # We gooien de error opnieuw, zodat de job stopt en de error zichtbaar is.
            raise

    print("--- Alle opgegeven kolommen zijn succesvol verwijderd, tabellen zijn vervangen en Identity-kolommen zijn bijgewerkt. ---")

#
# 8 PAS KOLOMVOLGORDE AAN
#


def pas_kolomvolgorde_aan(spark, catalog, target_schema, kolom_volgorde_mapping):
    """
    Past de volgorde van kolommen in Delta-tabellen aan op basis van een opgegeven mapping.

    Parameters:
    - spark: De SparkSession (vereist in Databricks).
    - catalog: De naam van de catalogus.
    - target_schema: De naam van het doel-schema.
    - kolom_volgorde_mapping: Een dictionary waarin de sleutels de tabelnamen zijn
      en de waarden lijsten zijn met de gewenste volgorde van kolommen.
    """
    for tabel, nieuwe_volgorde in kolom_volgorde_mapping.items():
        # Definieer de volledige namen van de tabellen
        oude_tabel_pad = f"{catalog}.{target_schema}.{tabel}"
        nieuwe_tabel_pad = f"{oude_tabel_pad}_temp_nieuwe_volgorde"

        print(f"Start verwerking voor tabel: {oude_tabel_pad}")

        try:
            # Lees de data van de originele tabel in een DataFrame
            df = spark.table(oude_tabel_pad)
            originele_kolommen = df.columns

            # Controleer of alle kolommen in de nieuwe volgorde bestaan in de originele tabel
            for kolom in nieuwe_volgorde:
                if kolom not in originele_kolommen:
                    raise ValueError(f"Kolom '{kolom}' bestaat niet in tabel '{tabel}'.")

            # Controleer op ontbrekende kolommen (exclusief SID-kolommen)
            ontbrekende_kolommen = [
                kolom for kolom in originele_kolommen
                if not kolom.lower().startswith("sid_") and kolom not in nieuwe_volgorde
            ]
            if ontbrekende_kolommen:
                raise ValueError(
                    f"De volgende kolommen ontbreken in de nieuwe volgorde voor tabel '{tabel}': {ontbrekende_kolommen}"
                )
            
            # 1. Bepaal de kolommen voor dataoverdracht (exclusief SID-kolommen)
            kolommen_om_te_selecteren = [
                kolom for kolom in nieuwe_volgorde
            ]
            
            # 2. Creëer een nieuw DataFrame met de correcte kolomvolgorde
            nieuwe_df = df.select(*kolommen_om_te_selecteren)
            
            print(f"Schrijft data naar tijdelijke tabel: {nieuwe_tabel_pad}")
            nieuwe_df.write.format("delta").mode("overwrite").saveAsTable(nieuwe_tabel_pad)

            # 3. Controleer of het aantal rijen overeenkomt
            originele_rijen_aantal = spark.table(oude_tabel_pad).count()
            nieuwe_rijen_aantal = spark.table(nieuwe_tabel_pad).count()
            if nieuwe_rijen_aantal != originele_rijen_aantal:
                raise ValueError(f"Aantal rijen in de nieuwe tabel ({nieuwe_rijen_aantal}) komt niet overeen met het aantal rijen in de originele tabel ({originele_rijen_aantal}).")

            print(f"Data succesvol overgezet naar tijdelijke tabel: {nieuwe_tabel_pad}")

            # 4. Verwijder de originele tabel en hernoem de tijdelijke
            print(f"Verwijdert originele tabel {oude_tabel_pad}...")
            spark.sql(f"DROP TABLE IF EXISTS {oude_tabel_pad}")

            print(f"Hernoemt {nieuwe_tabel_pad} naar {oude_tabel_pad}...")
            spark.sql(f"ALTER TABLE {nieuwe_tabel_pad} RENAME TO {oude_tabel_pad}")
            
            # 5. Vernieuw de metadata om de wijziging zichtbaar te maken
            print(f"Vernieuwt de catalogus metadata voor tabel {oude_tabel_pad}...")
            spark.catalog.refreshTable(oude_tabel_pad)

            print(f"Tabel {oude_tabel_pad} succesvol aangepast met nieuwe kolomvolgorde.")
            print("-" * 50)

        except Exception as e:
            print(f"Fout opgetreden bij verwerking van tabel {oude_tabel_pad}: {e}")
            raise

    print("De kolomvolgorde is succesvol aangepast voor alle opgegeven tabellen.")

#
# 9 UPDATE TABLE FROM OTHER TABLE
#


def update_table_from_other_table(spark, CATALOG, TARGET_SCHEMA, mapping):
    """
    Voert dynamische MERGE-updates uit in een Databricks Delta Lake-omgeving.
    """
    for map_item in mapping["mappings"]:
        source_table = map_item["source_table"]
        target_table = map_item["target_table"]
        source_fields = map_item["source_fields"]
        target_fields = map_item["target_fields"]
        id_field = map_item["id_field"]
        conditions = map_item.get("conditions", "1=1") 
        
        # 1. Definieer de SET-clausule (voor MERGE)
        # De syntax voor SET is eenvoudiger in MERGE (bijv. target.veld = source.veld)
        set_assignments = ", ".join(
            f"target.{target_field} = source.{source_field}"
            for source_field, target_field in zip(source_fields, target_fields)
        )
        
        # 2. Definieer de ON-clausule (de join-conditie)
        join_condition = f"target.{id_field} = source.{id_field}"
        
        # 3. Bouw de SQL-query met MERGE INTO
        query = f"""
        MERGE INTO {CATALOG}.{TARGET_SCHEMA}.{target_table} AS target
        USING {CATALOG}.{TARGET_SCHEMA}.{source_table} AS source
        ON {join_condition}
        WHEN MATCHED AND {conditions} THEN
            UPDATE SET {set_assignments}
        """
        
        # Voer de query uit
        spark.sql(query)
        print(f"Tabel {target_table} is geupdate met data van tabel {source_table}.")


# Databricks notebook source
import pytz
from datetime import datetime
from databricks.sdk.runtime import *
from pyspark.sql.types import *
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql import functions as F
from typing import *

from datateam_moss.logger import get_logger

logger = get_logger("__spark_io_utils__")

def sla_tabel_op_catalog(df: DataFrame, catalog: str, schema: str, tabel_naam: str, operatie: str, keuze: str):
    """
    Sla een DataFrame op als een tabel in een Databricks catalogus met specifieke opties.

    Parameters:
    df (DataFrame): De DataFrame die moet worden opgeslagen.
    catalog (str): De catalogus waarin de tabel moet worden opgeslagen.
    schema (str): Het schema waarin de tabel moet worden opgeslagen.
    tabel_naam (str): De naam van de tabel die moet worden opgeslagen.
    operatie (str): De operatie die moet worden uitgevoerd ('mergeSchema' of 'overwriteSchema').
    keuze (str): De keuze voor de operatie ('true' of 'false').

    Returns:
    None
    """
    if operatie not in ["mergeSchema", "overwriteSchema"]:
        raise ValueError("Ongeldige operatie. Kies uit 'mergeSchema' of 'overwriteSchema'.")
    
    if keuze not in ["true", "false"]:
        raise ValueError("Ongeldige keuze. Kies uit 'true' of 'false'.")

    options = {operatie: keuze}

    if operatie == "mergeSchema" and keuze == "true":
        df.write.option("mergeSchema", "true").mode("overwrite").saveAsTable(f"{catalog}.{schema}.{tabel_naam}")
    elif operatie == "mergeSchema" and keuze == "false":
        df.write.option("mergeSchema", "false").mode("overwrite").saveAsTable(f"{catalog}.{schema}.{tabel_naam}")
    elif operatie == "overwriteSchema" and keuze == "true":
        df.write.option("overwriteSchema", "true").mode("overwrite").saveAsTable(f"{catalog}.{schema}.{tabel_naam}")
    elif operatie == "overwriteSchema" and keuze == "false":
        df.write.option("overwriteSchema", "false").mode("overwrite").saveAsTable(f"{catalog}.{schema}.{tabel_naam}")


def del_meerdere_tabellen_catalog(catalog: str, schema: str, tabellen_filter: str, uitsluiten_tabellen: str = None):
    """
    Verwijder meerdere tabellen uit de catalogus.

    Args:
        catalog (str): Naam van de catalogus.
        schema (str): Naam van het schema.
        tabellen_filter (str): Filter voor tabellen die moeten worden verwijderd: "alles" of "prefix in tabelnaam"
        uitsluiten_tabellen (str, optional): Tabellen om uit te sluiten van verwijdering. Standaard is None.

    Raises:
        ValueError: Als er geen tabellen zijn die overeenkomen met het opgegeven filter.

    Returns:
        None
    """

    # Combineer catalogus en schema
    schema_catalog = f"{catalog}.{schema}"
    
    # Haal metadata op uit de Unity Catalog
    tabellen_catalog = spark.sql(f"SHOW TABLES IN {schema_catalog}")

    # Maak een set van alle tabellen in het opgegeven schema
    set_tabellen_catalog = {row["tableName"] for row in tabellen_catalog.collect()}

    # Filter de tabellen op basis van de gegeven filter
    if tabellen_filter == "alles":
        set_tabellen_catalog_filter = set_tabellen_catalog
    else:
        set_tabellen_catalog_filter = {table for table in set_tabellen_catalog if tabellen_filter in table}

    # Verwijder de tabellen die in uitsluiten_tabellen staan, als deze parameter is meegegeven
    if uitsluiten_tabellen:
        uitsluiten_set = set(uitsluiten_tabellen.split(","))
        set_tabellen_catalog_filter -= uitsluiten_set

    # Controleer of er tabellen zijn die voldoen aan het opgegeven filter
    if not set_tabellen_catalog_filter:
        raise ValueError("Er bestaan geen tabellen met het opgegeven tabellen_filter. Vul de parameter uitsluiten_tabellen aan of geef een ander filter op.")
    
    # Print de geselecteerde tabellen
    print("Geselecteerde tabellen voor verwijdering:", set_tabellen_catalog_filter)
    
    # Vraag om bevestiging voor verwijdering van de tabellen
    verwijder_check = input("Je staat op het punt om deze tabellen te verwijderen uit de CATALOG. Typ 'ja' om de tabellen te verwijderen -> ")
    
    # Verwijder de geselecteerde tabellen indien bevestigd
    if verwijder_check.lower() == "ja":
        for table in set_tabellen_catalog_filter:
            spark.sql(f"DROP TABLE {schema_catalog}.{table}")
        print("De opgegeven tabellen zijn correct verwijderd.") 
    else:
        print("Verwijdering geannuleerd.")
    return

def schrijf_naar_metatabel(catalogus: str, meta_schema: str, tabel_schema: str, tabel_naam: str, script_naam: str, controle: str, controle_waarde: str, meta_tabel_naam: str):
    """
    Schrijft gegevens naar de meta-tabel in de opgegeven catalogus en schema.

    Args:
        catalogus (str): De naam van de catalogus waarin de meta-tabel zich bevindt.
        meta_schema (str): De naam van het schema waarin de meta-tabel zich bevindt.
        tabel_schema (str): De naam van het schema van de tabel waarvoor de gegevens worden toegevoegd.
        tabel_naam (str): De naam van de tabel waarvoor de gegevens worden toegevoegd.
        script_naam (str): De naam van het script dat de gegevens toevoegt.
        controle (str): Informatie over de controle.
        controle_waarde (str): De waarde van de controle.
        meta_tabel_naam (str): De naam van de meta-tabel waar gegevens worden opgeslagen.

    Returns:
        None
    """
    try: 
        meta_tabel_df = spark.read.table(f"{catalogus}.{meta_schema}.{meta_tabel_naam}")
        data = [(catalogus, tabel_schema, tabel_naam, script_naam, controle, controle_waarde)]
        kolommen = ["table_catalog", "table_schema", "table_name", "script", "controle", "controle_waarde"]
        temp_tabel = spark.createDataFrame(data, kolommen)
        union_df = meta_tabel_df.union(temp_tabel)

        updated_df = (union_df
                    .select([F.when(F.col(c) == "", None).otherwise(F.col(c)).alias(c) for c in union_df.columns])
                    .na.drop(how="all").distinct())

        updated_df.write.saveAsTable(f"{catalogus}.{meta_schema}.{meta_tabel_naam}", mode="overwrite")
    
    except:
        data = [(catalogus, tabel_schema, tabel_naam, script_naam, controle, controle_waarde)]
        kolommen = ["table_catalog", "table_schema", "table_name", "script", "controle", "controle_waarde"]
        temp_tabel = spark.createDataFrame(data, kolommen)
        temp_tabel.write.saveAsTable(f"{catalogus}.{meta_schema}.{meta_tabel_naam}", mode="overwrite")

def convert_datetime_format(input_format):
    """
    Converteert het opmaakoptie voor datum en tijd van het datetime (format)  naar het PySpark format.
    
    Args:
        input_format (str): De invoeropmaakoptie voor datum en tijd.
    
    Returns:
        str: De geconverteerde opmaakoptie voor datum en tijd.
    """
    # Mapt de formatting van de package: datetime -> PySpark formatting
    format_mapping = {
        "%Y": "yyyy",
        "%m": "MM",
        "%d": "dd",
        "%H": "HH",
        "%M": "mm",
        "%S": "ss"
    }

    # Vervang opmaaktekens in de opgegeven string
    for char, replacement in format_mapping.items():
        input_format = input_format.replace(char, replacement)
    
    return(input_format)

def tijdzone_amsterdam(tijdformaat="%Y-%m-%d %H:%M:%S", date_string_timestamp="timestamp"):
    """Haalt de huidige tijd op en converteert deze naar het opgegeven tijdsformaat en de tijdzone van Amsterdam.
    """
    converted_format = convert_datetime_format(tijdformaat)
    amsterdam_tz = pytz.timezone('Europe/Amsterdam')
    huidige_datum = datetime.now(amsterdam_tz).strftime(tijdformaat)

    if date_string_timestamp == "string":
        return huidige_datum
    elif date_string_timestamp == "timestamp":
        return F.to_timestamp(F.lit(huidige_datum), converted_format)
    elif date_string_timestamp == "date":
        return F.to_date(F.lit(huidige_datum), converted_format)

def vind_scheidingsteken(bestandspad, scheidingstekens=[',', ';', '\t', '|']):
    """
    Probeert het juiste scheidingsteken (delimiter) te vinden voor een CSV-bestand door verschillende mogelijke scheidingstekens te testen.

    Deze functie leest het CSV-bestand met elk gegeven scheidingsteken en controleert of het resulterende DataFrame
    meer dan één kolom bevat. Indien ja, wordt aangenomen dat het gevonden scheidingsteken correct is.

    Args:
        bestandspad (str): Het pad naar het CSV-bestand in the UC.
        scheidingstekens (list): Lijst van mogelijke scheidingstekens om te testen. Standaard zijn dit: [',', ';', '\t', '|'].
    """
    for scheidingsteken in scheidingstekens:
        try:
            df = spark.read.csv(bestandspad, sep=scheidingsteken, header=True)
            if len(df.columns) > 1:
                return scheidingsteken, df.columns
        except Exception:
            continue
    return None, None

def create_table_from_ddl(
    spark: SparkSession, 
    table_definition: Dict[str, Any], 
    full_table_name: str,
    genereer_sid: bool = False
) -> None:
    """
    Maakt een Delta-tabel aan in Spark op basis van een DDL-definitie, tenzij de tabel al bestaat.

    Args:
        spark (SparkSession): De actieve SparkSession.
        table_definition (Dict[str, Any]): 
            Dictionary met minimaal een "columns"-sleutel. 
            "columns" moet een lijst bevatten van kolomdefinities in het formaat:
            - name (str): Naam van de kolom
            - type (str): Spark datatype als string, bv. "StringType()"
            - nullable (bool): Of de kolom null-waarden toestaat
        full_table_name (str): Volledig gekwalificeerde tabelnaam, bijv. "catalog.schema.table".
        genereer_sid (bool, optional): Of een "sid" kolom moet worden gegenereerd. Defaults is False.
        voorbeeld_table_def = {
            "tabel_1_naam": {
            "columns": [
            {"name": "kolom1", "type": "IntegerType()", "nullable": True},
            {"name": "kolom2", "type": "IntegerType()", "nullable": True}
            ]
            }
        }
 
    Returns:
        None
    """

    if spark.catalog.tableExists(full_table_name):
        print(f"Tabel {full_table_name} bestaat al. Er wordt niets gedaan.")
        return
    
    # Parse tabelnaam
    table_name_only = full_table_name.split(".")[-1]
    entity_name = table_name_only.split("_")[1]

    # Bouw kolommenlijst
    columns_ddl = []
    for col in table_definition["columns"]:
        col_def = f"`{col['name']}` {col['type'].replace('Type()', '').replace('ArrayType(String)', 'Array<string>').upper()}"
        if not col.get("nullable", True):
            col_def += " NOT NULL"
        columns_ddl.append(col_def)
    
    # Eventueel SID kolom toevoegen
    if genereer_sid:
        sid_col = f"`sid_{entity_name}` BIGINT GENERATED ALWAYS AS IDENTITY (START WITH 1 INCREMENT BY 1)"
        columns_ddl.insert(0, sid_col)  # zet SID vooraan
    
    # Combineer alles in CREATE TABLE DDL
    ddl = f"""
    CREATE TABLE {full_table_name} (
        {', '.join(columns_ddl)}
    )
    USING DELTA
    """
    print (ddl)
    # Maak de tabel aan
    spark.sql(ddl)
    print(f"Tabel {full_table_name} is aangemaakt{' met SID' if genereer_sid else ''}.")

def create_stringtype_dataframe_from_list(spark, data: List) -> DataFrame:
    """
    Converteert een lijst van dictionaries naar een PySpark DataFrame, waarbij alles als string wordt opgeslagen om de data zo ruw mogelijk op te halen.

    Parameters:
        spark (SparkSession): De actieve SparkSession.
        data (list): Lijst met data om in de DataFrame te laden.
    
    Returns:
        DataFrame: Een PySpark DataFrame met het opgegeven schema en data.
    """

    if not isinstance(data, list):
        raise TypeError("Parameter 'data' moet een lijst zijn.")

    if len(data) == 0:
        raise ValueError("Parameter 'data' mag niet leeg zijn.")

    schema = StructType([
    StructField(col, StringType(), True) 
    for col in data[0].keys()
        ])

    try:
        df = spark.createDataFrame(data, schema=schema)
    except Exception as e:
        raise Exception(f'Fout bij het converteren van de data naar een PySpark DataFrame: {e}')

    return df

def add_metadata_columns_to_dataframe(df: DataFrame, m_columns: List, runtime: datetime, bron: str = None) -> DataFrame:
    """
    Voegt metadata-kolommen toe aan een PySpark DataFrame.

    Parameters:
        df (DataFrame): De PySpark DataFrame waarop de metadata-kolommen moeten worden toegevoegd.
        m_columns (List): Een lijst met kolomnamen die moeten worden toegevoegd als metadata-kolommen. 
        runtime (str): De runtime van de notebook
        bron (str): De bron van de data

    Returns:
        DataFrame: Een PySpark DataFrame met de toegevoegde metadata-kolommen.
    """
    runtime_str = runtime.strftime("%Y%m%d%H%M%S")

    if not isinstance(m_columns, list):
        raise TypeError("Parameter 'columns' moet een lijst zijn.")
    if not isinstance(runtime, datetime):
        raise TypeError("Parameter 'runtime' moet een datetime object zijn.")

    if "m_geldig_van" in m_columns:
        df = df.withColumn("m_geldig_van", F.to_timestamp(F.lit("1900-12-31")))
    if "m_geldig_tot" in m_columns:
        df = df.withColumn("m_geldig_tot", F.to_timestamp(F.lit("9000-12-31")))
    if "m_is_actief" in m_columns:
        df = df.withColumn("m_is_actief", F.lit(True))
    if "m_aangemaakt_op" in m_columns:
        df = df.withColumn("m_aangemaakt_op", F.to_timestamp(F.lit(runtime)))
    if "m_gewijzigd_op" in m_columns:
        df = df.withColumn("m_gewijzigd_op", F.to_timestamp(F.lit(runtime)))
    if "m_bron" in m_columns:
        df = df.withColumn("m_bron", F.lit(bron))
    if "m_runid" in m_columns:
        df = df.withColumn("m_runid", F.lit(runtime_str))
    
    return df

def write_to_table(
    df: DataFrame, 
    target_table: str,
    mode: str = "overwrite",
    schema_evolution_strategy: Dict[str, str] = {"mergeSchema": "false"}
) -> None:
    """
    Schrijft een dataframe naar een opgegeven tabel in Spark met schema-evolutie-opties.

    Parameters:
        df (DataFrame): Het dataframe dat moet worden weggeschreven.
        target_table (str): De volledige naam van de doel-tabel inc. catalog en schema.
        mode (str): De schrijfmethode (bijv. "overwrite", "append", "error", "ignore").
                    Standaard is "overwrite".
        schema_evolution_strategy (Dict[str, str]): Opties voor schema-evolutie, zoals {"mergeSchema": "false"}.
                    Standaard bevat het {"mergeSchema": "false"}.

    Raises:
        ValueError: Als de target_table niet is opgegeven.
        Exception: Als er een fout optreedt tijdens het schrijven.
    """
    if not target_table:
        raise ValueError("De target_table is niet opgegeven. Schrijven wordt gestopt.")

    # Zorg ervoor dat "mergeSchema" standaard op "false" staat als het niet is meegegeven
    schema_evolution_strategy.setdefault("mergeSchema", "false")

    try:
        logger.info(f"Start schrijven naar tabel {target_table} met mode '{mode}' en schema-evolutie {schema_evolution_strategy}.")

        # Pas de schema-evolutie-opties toe en schrijf het dataframe naar de tabel
        writer = df.write
        for key, value in schema_evolution_strategy.items():
            writer = writer.option(key, value)
        
        writer.mode(mode).saveAsTable(target_table)
        
        logger.info(f"Data succesvol geschreven naar tabel {target_table}.")
    except Exception as e:
        logger.error(f"Fout bij het schrijven naar tabel {target_table}: {e}")
        raise  # Laat de fout opnieuw opgooien om de pipeline te laten stoppen


"""Logging module for the OpenTelemetry integration that sends logs to log analytics workspace
and adds a console logger for immediate feedback during development.
"""
import sys
from pyspark.dbutils import DBUtils
from pyspark.sql import SparkSession
import logging
from azure.monitor.opentelemetry.exporter import AzureMonitorLogExporter
from opentelemetry.sdk._logs import (
    LoggerProvider,
    LoggingHandler,
)
from opentelemetry._logs import set_logger_provider, get_logger_provider
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor

spark = SparkSession.builder.getOrCreate()
dbutils = DBUtils(spark)

def get_logger(name: str,
               stream_log_format: str = '%(asctime)s - %(levelname)s - %(message)s',
               azure_log_format: str = '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
               app_insights_connection_secret: str = 'app-insights-connection') -> logging.Logger:
    """Create and configure a logger with:
    - a `StreamHandler` that writes to `stdout`/notebook outputs using `stream_log_format`
    - an Azure Monitor exporter, initialized once globally, using a connection string retrieved from Azure Key Vault.
    - an `LoggingHandler` that forwards logs to Azure Monitor using `azure_log_format` (level `WARN`). Handlers are added only if they
      are not already present to avoid duplicate log entries.

    Parameters
    ----------
    name : str
        Name of the logger.
    stream_log_format : str, optional
        Format for console log messages. Defaults to '%(asctime)s - %(levelname)s - %(message)s'.
    azure_log_format : str, optional
        Format for Azure Monitor log messages. Defaults to '%(name)s - %(levelname)s - %(message)s'.
    app_insights_secret : str, optional
        Secret name in Key Vault that holds the Azure Monitor connection string.

    Returns
    -------
    logging.Logger
        The configured logger instance.
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    logger.propagate = False  # prevent duplicate logs from root logger

    # --- Set StreamHandler ---
    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(logging.INFO)
        stream_handler.setFormatter(logging.Formatter(stream_log_format))
        logger.addHandler(stream_handler)

    # --- Initialize Azure Monitor backend ONCE globally ---
    current_provider = get_logger_provider()
    if not isinstance(current_provider, LoggerProvider):
        try:
            connection_string = dbutils.secrets.get(
                scope='keyvault', key=app_insights_connection_secret)
        except Exception as e:
            logger.error(f"Failed to retrieve secret: {e}")
            connection_string = None

        if connection_string:
            try:
                exporter = AzureMonitorLogExporter(
                    connection_string=connection_string)
                logger_provider = LoggerProvider()
                logger_provider.add_log_record_processor(
                    BatchLogRecordProcessor(exporter))
                set_logger_provider(logger_provider)
            except Exception as e:
                logger.error(f"Failed to initialize Azure Monitor exporter: {e}")

        else:
            logger.error(
                "AzureMonitorLogExporter not initialized due to missing connection string.")

    # --- Attach Azure handler to this logger (only once) ---
    if not any(isinstance(h, LoggingHandler) for h in logger.handlers):
        try:
            azure_handler = LoggingHandler()
            azure_handler.setLevel(logging.WARN)
            azure_handler.setFormatter(logging.Formatter(azure_log_format))
            logger.addHandler(azure_handler)
        except Exception as e:
            logger.warning(f"Failed to attach Azure LoggingHandler: {e}")

    return logger


import pytest
import unittest
from datetime import datetime

from pyspark.sql import SparkSession
from pyspark.sql.types import StringType

from datateam_moss import spark_io_utils

spark = SparkSession.builder.getOrCreate()

class TestGetData(unittest.TestCase):

    def test_create_stringtype_dataframe_from_list_basic(self):
        data = [
            {"name": "Alice", "age": "30"},
            {"name": "Bob", "age": "25"}
        ]

        df = spark_io_utils.create_stringtype_dataframe_from_list(spark, data)

        # kolomnamen moeten overeenkomen
        assert df.columns == ["name", "age"]

        # datatypes moeten allemaal StringType zijn
        for field in df.schema.fields:
            assert isinstance(field.dataType, StringType)

        # data moet correct geladen worden
        result = df.collect()
        assert result[0]["name"] == "Alice"
        assert result[1]["age"] == "25"


    def test_create_stringtype_dataframe_from_list_string_conversion(self):
        data = [
            {"id": 1, "price": 99.5},
            {"id": 2, "price": 100.0}
        ]

        df = spark_io_utils.create_stringtype_dataframe_from_list(spark, data)

        # alles moet strings worden
        result = df.collect()
        assert result[0]["id"] == "1"
        assert result[0]["price"] == "99.5"

    def test_create_stringtype_dataframe_from_list_empty_list(self):
        data = []

        with pytest.raises(ValueError):
            spark_io_utils.create_stringtype_dataframe_from_list(spark, data)

    def test_create_stringtype_dataframe_from_list_no_list(self):
        data = {}

        with pytest.raises(TypeError):
            spark_io_utils.create_stringtype_dataframe_from_list(spark, data)
        
    def test_add_metadata_columns_to_dataframe_check_type(self):
        data = [
            {"id": 1, "price": 99.5},
            {"id": 2, "price": 100.0}
        ]

        df = spark.createDataFrame(data=data, schema=["id", "price"])

        with pytest.raises(TypeError):
            spark_io_utils.add_metadata_columns_to_dataframe(df=df, m_columns={"m_bron"}, runtime=datetime.now(), bron= "test")

    def test_add_metadata_columns_to_dataframe_runtime_and_timestamps(self):
        data = [
            {"id": 1, "price": 99.5},
            {"id": 2, "price": 100.0}
        ]

        df = spark.createDataFrame(data=data, schema=["id", "price"])

        runtime = datetime(2023, 5, 10, 8, 45)

        result = spark_io_utils.add_metadata_columns_to_dataframe(
            df,
            ["m_aangemaakt_op"],
            runtime,
            "test_bron"
        )

        # timestamp converted correct?
        first_row = result.first()
        assert str(first_row["m_aangemaakt_op"]).startswith("2023-05-10 08:45")
            

if __name__ == '__main__':
    import sys
    unittest.main(argv=['first-arg-is-ignored'], exit=False)